/** @file scim_hangul_imengine.cpp
 */

/*
 * Smart Common Input Method
 * 
 * Copyright (C) 2004-2006 Choe Hwanjin
 * Copyright (c) 2004-2006 James Su <suzhe@tsinghua.org.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: scim_hangul_imengine.cpp,v 1.8 2006/02/05 15:17:17 hwanjin Exp $
 */

#define Uses_SCIM_UTILITY
#define Uses_SCIM_IMENGINE
#define Uses_SCIM_LOOKUP_TABLE
#define Uses_SCIM_CONFIG_BASE

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#include <scim.h>
#include "scim_hangul_private.h"
#include "scim_hangul_imengine.h"

#ifdef HAVE_GETTEXT
  #include <libintl.h>
  #define _(String) dgettext(GETTEXT_PACKAGE,String)
  #define N_(String) (String)
#else
  #define _(String) (String)
  #define N_(String) (String)
  #define bindtextdomain(Package,Directory)
  #define textdomain(domain)
  #define bind_textdomain_codeset(domain,codeset)
#endif

#define scim_module_init hangul_LTX_scim_module_init
#define scim_module_exit hangul_LTX_scim_module_exit
#define scim_imengine_module_init hangul_LTX_scim_imengine_module_init
#define scim_imengine_module_create_factory hangul_LTX_scim_imengine_module_create_factory


#ifndef SCIM_HANGUL_ICON_FILE
    #define SCIM_HANGUL_ICON_FILE           (SCIM_ICONDIR "/scim-hangul.png")
#endif

#define SCIM_PROP_PREFIX				"/IMEngine/Hangul"
#define SCIM_PROP_INPUT_MODE				SCIM_PROP_PREFIX "/InputMode"
#define SCIM_PROP_INPUT_MODE_LATIN			SCIM_PROP_INPUT_MODE "/Latin"
#define SCIM_PROP_INPUT_MODE_HANGUL			SCIM_PROP_INPUT_MODE "/Hangul"
#define SCIM_PROP_INPUT_LAYOUT				SCIM_PROP_PREFIX "/InputLayout"
#define SCIM_PROP_INPUT_LAYOUT_2BUL			SCIM_PROP_INPUT_LAYOUT "/2Bul"
#define SCIM_PROP_INPUT_LAYOUT_3BUL_2BUL_SHIFTED	SCIM_PROP_INPUT_LAYOUT "/3Bul2BulShifted"
#define SCIM_PROP_INPUT_LAYOUT_3BUL_390			SCIM_PROP_INPUT_LAYOUT "/3Bul390"
#define SCIM_PROP_INPUT_LAYOUT_3BUL_FINAL		SCIM_PROP_INPUT_LAYOUT "/3BulFinal"
#define SCIM_PROP_INPUT_LAYOUT_3BUL_NO_SHIFT		SCIM_PROP_INPUT_LAYOUT "/3BulNoShift"
#define SCIM_PROP_INPUT_LAYOUT_3BUL_YETGEUL		SCIM_PROP_INPUT_LAYOUT "/3BulYetgeul"


struct CandidateItem {
    ucs4_t      ch;
    const char *comment;
};

#include "candidatetable.h"

static ConfigPointer _scim_config (0);

extern "C" {
    void scim_module_init (void)
    {
        bindtextdomain (GETTEXT_PACKAGE, SCIM_HANGUL_LOCALEDIR);
        bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
    }

    void scim_module_exit (void)
    {
        _scim_config.reset ();
    }

    uint32 scim_imengine_module_init (const ConfigPointer &config)
    {
        SCIM_DEBUG_IMENGINE(1) << "Initialize Hangul Engine\n";

        _scim_config = config;

        return 1;
    }

    IMEngineFactoryPointer scim_imengine_module_create_factory (uint32 engine)
    {
        HangulFactory *factory = 0;

        try {
            factory = new HangulFactory (_scim_config,
					 "d75857a5-4148-4745-89e2-1da7ddaf70a9");
        } catch (...) {
            delete factory;
            factory = 0;
        }

        return factory;
    }
}

HangulFactory::HangulFactory (const ConfigPointer     &config,
			      const char              *uuid)
    : m_uuid                   (uuid),
      m_keyboard_type          (HANGUL_KEYBOARD_2),
      m_show_candidate_comment (true),
      m_use_ascii_mode         (SCIM_CONFIG_IMENGINE_HANGUL_USE_ASCII_MODE_DEFAULT),
      m_hide_layout            (SCIM_CONFIG_IMENGINE_HANGUL_HIDE_LAYOUT_DEFAULT),
      m_config                 (config)
{
    set_languages ("ko");

    reload_config (m_config);
    m_reload_signal_connection = m_config->signal_connect_reload (slot (this, &HangulFactory::reload_config));
}

HangulFactory::~HangulFactory ()
{
    m_reload_signal_connection.disconnect ();
}

WideString
HangulFactory::get_name () const
{
	return utf8_mbstowcs (String ("Hangul"));
}

WideString
HangulFactory::get_authors () const
{
    return WideString ();
}

WideString
HangulFactory::get_credits () const
{
    return WideString ();
}

WideString
HangulFactory::get_help () const
{
    const char *header =
	_("Key bindings:\n\n");
    const char *paragraph1 =
	_("  On/Off toggle key: Alt_R, Shift+space\n"
	  "    This key binding is to switch the input mode between the ASCII input\n"
	  "    mode and the hangul input.\n"
	  "    on the default key bindings, it will be effected by tapping\n"
	  "    the right Alt key or press the Space key with holding the Shift key.\n\n");
    const char *paragraph2 =
	_("  Hangul Hanja key: Control_R, F9\n"
	  "    This key binding is to convert a hangul character to a hanja character.\n"
	  "    it will be effected by tapping the right Control key or press F9 key.\n\n");

    return utf8_mbstowcs (header)
	+ utf8_mbstowcs (paragraph1)
	+ utf8_mbstowcs (paragraph2);
}

String
HangulFactory::get_uuid () const
{
    return m_uuid;
}

String
HangulFactory::get_icon_file () const
{
    return String (SCIM_HANGUL_ICON_FILE);
}

IMEngineInstancePointer
HangulFactory::create_instance (const String &encoding, int id)
{
    SCIM_DEBUG_IMENGINE(1) << "create_instance: HangulInstance.\n";
    return new HangulInstance (this, encoding, id);
}

void
HangulFactory::append_config_listener (HangulInstance *listener)
{
    bool found = false;
    std::vector<HangulInstance*>::iterator it;
    for (it = m_config_listeners.begin();
         it != m_config_listeners.end();
         it++)
    {
        if (*it == listener) {
            found = true;
            break;
        }
    }

    if (!found)
        m_config_listeners.push_back (listener);
}

void
HangulFactory::remove_config_listener (HangulInstance *listener)
{
    std::vector<HangulInstance*>::iterator it;
    for (it = m_config_listeners.begin();
         it != m_config_listeners.end();
         it++)
    {
        if (*it == listener) {
            m_config_listeners.erase (it);
            break;
        }
    }
}

void
HangulFactory::reload_config (const ConfigPointer &config)
{
    if (config) {
        m_show_candidate_comment =
            config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_SHOW_CANDIDATE_COMMENT),
                          m_show_candidate_comment);

	m_use_ascii_mode =
            config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_USE_ASCII_MODE),
                          m_use_ascii_mode);

	m_hide_layout =
            config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_HIDE_LAYOUT),
                          m_hide_layout);

        String str;

        str = config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_HANGUL_HANJA_KEY),
                            String (SCIM_CONFIG_IMENGINE_HANGUL_HANGUL_HANJA_KEY_DEFAULT));

        scim_string_to_key_list (m_hangul_hanja_keys, str);

	str = config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_ON_OFF_KEY),
			    String (SCIM_CONFIG_IMENGINE_HANGUL_ON_OFF_KEY_DEFAULT));
	scim_string_to_key_list (m_on_off_keys, str);

	m_input_layout = config->read (String (SCIM_CONFIG_IMENGINE_HANGUL_INPUT_LAYOUT),
				       String (SCIM_CONFIG_IMENGINE_HANGUL_INPUT_LAYOUT_DEFAULT));
	/* it may be the wrong assumption that it may be false
	 * if no /Panel/Gtk/LookupTableVertical defined. any APIs to get it may be needed?
	 */
	m_lookup_table_vertical = config->read (String (SCIM_CONFIG_PANEL_LOOKUP_TABLE_VERTICAL),
						false);
    }

    if (!m_hangul_hanja_keys.size ()) {
        m_hangul_hanja_keys.push_back (KeyEvent (SCIM_KEY_Hangul_Hanja, 0));
        m_hangul_hanja_keys.push_back (KeyEvent (SCIM_KEY_F9, 0));
    }
    if (!m_on_off_keys.size ()) {
        m_on_off_keys.push_back (KeyEvent (SCIM_KEY_Alt_R, SCIM_KEY_ReleaseMask));
    }

    /* reload config for all instances */
    std::vector<HangulInstance*>::iterator it;

    for (it = m_config_listeners.begin ();
	 it != m_config_listeners.end ();
	 it++) {
	(*it)->reload_config (config);
    }
}

HangulInstance::HangulInstance (HangulFactory *factory,
                                const String  &encoding,
                                int            id)
    : IMEngineInstanceBase (factory, encoding, id),
      m_factory (factory),
      m_prev_key (0,0),
      m_input_mode (INPUT_MODE_HANGUL),
      m_output_mode (OUTPUT_MODE_SYLLABLE),
      m_on_init (true),
      m_input_layout (SCIM_HANGUL_INPUT_LAYOUT_2BUL)
{
    m_factory->append_config_listener (this);
    m_hic = hangul_ic_new(factory->m_keyboard_type);
    hangul_ic_reset(m_hic);
    reload_config (m_factory->m_config);
    m_on_init = false;

    char label[16];
    std::vector <WideString> labels;

    for (int i = 1; i < 10; ++i) {
        snprintf (label, sizeof(label), "%d", i);
        labels.push_back (utf8_mbstowcs (label));
    }

    m_lookup_table.set_candidate_labels (labels);
}

HangulInstance::~HangulInstance ()
{
    m_factory->remove_config_listener (this);
}

void
HangulInstance::reload_config (const ConfigPointer &config)
{
    if (m_on_init) {
        if (m_factory->m_input_layout == "2bul")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_2BUL);
        else if (m_factory->m_input_layout == "3bul2bulShifted")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_2BUL_SHIFTED);
        else if (m_factory->m_input_layout == "3bul390")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_390);
        else if (m_factory->m_input_layout == "3bulFinal")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_FINAL);
        else if (m_factory->m_input_layout == "3bulNoShift")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_NO_SHIFT);
        else if (m_factory->m_input_layout == "3bulYetgeul")
            set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_YETGEUL);

	if (m_factory->m_use_ascii_mode)
	    set_input_mode (INPUT_MODE_DIRECT);
	else
	    set_input_mode (INPUT_MODE_HANGUL);
    }

    m_properties.clear ();
    install_properties ();
}

void
HangulInstance::cursor_up_on_lookup_table(void)
{
	m_lookup_table.cursor_up ();
	update_lookup_table (m_lookup_table);
	hangul_update_aux_string ();
}

void
HangulInstance::cursor_down_on_lookup_table(void)
{
	m_lookup_table.cursor_down ();
	update_lookup_table (m_lookup_table);
	hangul_update_aux_string ();
}

void
HangulInstance::page_up_on_lookup_table(void)
{
	m_lookup_table.page_up ();
	update_lookup_table (m_lookup_table);
	hangul_update_aux_string ();
}

void
HangulInstance::page_down_on_lookup_table(void)
{
	m_lookup_table.page_down ();
	update_lookup_table (m_lookup_table);
	hangul_update_aux_string ();
}

bool
HangulInstance::candidate_key_event (const KeyEvent &key)
{
    switch (key.code) {
        case SCIM_KEY_Return:
        case SCIM_KEY_KP_Enter:
            select_candidate (m_lookup_table.get_cursor_pos_in_current_page ());
            break;
        case SCIM_KEY_BackSpace:
	    hangul_ic_backspace(m_hic);
            hangul_update_preedit_string ();
	    update_candidates ();
            break;
        case SCIM_KEY_KP_Subtract:
            cursor_up_on_lookup_table();
            break;
        case SCIM_KEY_Left:
        case SCIM_KEY_h:
            if (m_factory->m_lookup_table_vertical)
                page_up_on_lookup_table();
            else
                cursor_up_on_lookup_table();
            break;
        case SCIM_KEY_space:
        case SCIM_KEY_KP_Add:
            cursor_down_on_lookup_table();
            break;
        case SCIM_KEY_Right:
        case SCIM_KEY_l:
            if (m_factory->m_lookup_table_vertical)
                page_down_on_lookup_table();
            else
                cursor_down_on_lookup_table();
            break;
        case SCIM_KEY_Page_Up:
            page_up_on_lookup_table();
            break;
        case SCIM_KEY_Up:
        case SCIM_KEY_k:
            if (m_factory->m_lookup_table_vertical)
                cursor_up_on_lookup_table();
            else
                page_up_on_lookup_table();
            break;
        case SCIM_KEY_Page_Down:
        case SCIM_KEY_KP_Tab:
            page_down_on_lookup_table();
            break;
        case SCIM_KEY_Down:
        case SCIM_KEY_j:
            if (m_factory->m_lookup_table_vertical)
                cursor_down_on_lookup_table();
            else
                page_down_on_lookup_table();
            break;
        case SCIM_KEY_Escape:
            delete_candidates ();
            break;
        case SCIM_KEY_1: 
        case SCIM_KEY_2: 
        case SCIM_KEY_3: 
        case SCIM_KEY_4: 
        case SCIM_KEY_5: 
        case SCIM_KEY_6: 
        case SCIM_KEY_7: 
        case SCIM_KEY_8: 
        case SCIM_KEY_9: 
            select_candidate (key.code - SCIM_KEY_1);
            break;
        default:
            break;
    }

    return true;
}

bool
HangulInstance::process_key_event (const KeyEvent& rawkey)
{
    SCIM_DEBUG_IMENGINE(1) << "process_key_event.\n";
    KeyEvent key = rawkey.map_to_layout(SCIM_KEYBOARD_Default);

    m_prev_key = key;

    /* process on/off keys here since it allows you to
     * activate/inactivate IME with key releasee event
     * such as tapping alt key
     */
    if (is_on_off_key (key)) {
        toggle_on_off_state();
        return true;
    }

    /* ignore key release. */
    if (key.is_key_release ())
        return false;

    if (!is_ime_activated ()) {
        /* ASCII mode */
        return false;
    }

    if (is_backspace_key(key)) {
        bool ret = hangul_ic_backspace(m_hic);
        if (ret) {
	    if (m_lookup_table.number_of_candidates ()) {
		WideString wstr = get_preedit_string();
		if (wstr.empty())
		    delete_candidates ();
		else
		    update_candidates ();
	    }
	    hangul_update_preedit_string ();
	}
        return ret;
    }

    /* toggle candidate table */
    if (is_hangul_hanja_key (key)) {
        if (m_lookup_table.number_of_candidates ())
            delete_candidates ();
        else
            update_candidates ();

        return true;
    }

    /* ignore shift keys */
    if (key.code == SCIM_KEY_Shift_L || key.code == SCIM_KEY_Shift_R)
        return false;

    /* reset on modifier-on keys */
    if (key.is_control_down() || key.is_alt_down()) {
	reset ();
        return false;
    }

    /* candidate keys */
    if (m_lookup_table.number_of_candidates ()) {
        return candidate_key_event (key);
    }

    if (key.code >= SCIM_KEY_exclam && key.code <= SCIM_KEY_asciitilde) {
	/* main hangul composing process */
	int ascii = key.get_ascii_code();
	if (key.is_caps_lock_down()) {
	    if (isupper(ascii))
		ascii = tolower(ascii);
	    else if (islower(ascii))
		ascii = toupper(ascii);
	}

	bool ret = hangul_ic_filter(m_hic, ascii);

	WideString wstr;
	wstr = get_commit_string ();
	if (wstr.length ()) {
	    /* Before commit, we set preedit string to null to work arround
	     * some buggy IM implementation, ex) Qt, Evolution */
	    hide_preedit_string ();
	    commit_string(wstr);
	}

	hangul_update_preedit_string ();

	return ret;
    }

    reset ();
    return false;
}

void
HangulInstance::move_preedit_caret (unsigned int pos)
{
}

void
HangulInstance::select_candidate (unsigned int index)
{
    SCIM_DEBUG_IMENGINE(2) << "select_candidate.\n";

    if ((int)index < m_lookup_table.get_current_page_size ()) {
        WideString wstr = m_lookup_table.get_candidate_in_current_page (index);
        hangul_ic_reset (m_hic);
        commit_string (wstr);
        delete_candidates ();
        hangul_update_preedit_string ();
    }
}

void
HangulInstance::update_lookup_table_page_size (unsigned int page_size)
{
    SCIM_DEBUG_IMENGINE(2) << "update_lookup_table_page_size.\n";

    m_lookup_table.set_page_size (page_size);
}

void
HangulInstance::lookup_table_page_up ()
{
    if (!m_lookup_table.number_of_candidates () || !m_lookup_table.get_current_page_start ())
        return;

    SCIM_DEBUG_IMENGINE(2) << "lookup_table_page_up.\n";

    m_lookup_table.page_up ();

    update_lookup_table (m_lookup_table);

    hangul_update_aux_string ();
}

void
HangulInstance::lookup_table_page_down ()
{
    if (m_lookup_table.number_of_candidates () <= 0 ||
        m_lookup_table.get_current_page_start () + m_lookup_table.get_current_page_size () >=
          (int)m_lookup_table.number_of_candidates ())
        return;

    SCIM_DEBUG_IMENGINE(2) << "lookup_table_page_down.\n";

    m_lookup_table.page_down ();

    update_lookup_table (m_lookup_table);

    hangul_update_aux_string ();
}

void
HangulInstance::reset ()
{
    SCIM_DEBUG_IMENGINE(2) << "reset.\n";

    hangul_ic_reset(m_hic);

    hide_preedit_string();

    WideString wstr = get_commit_string();
    if (wstr.length())
        commit_string(wstr);

    delete_candidates ();
}

void
HangulInstance::focus_in ()
{
    SCIM_DEBUG_IMENGINE(2) << "focus_in.\n";

    if (m_lookup_table.number_of_candidates ()) {
        update_lookup_table (m_lookup_table);
        show_lookup_table ();
    } else {
        hide_lookup_table ();
    }

    hangul_update_aux_string ();

    install_properties ();
}

void
HangulInstance::focus_out ()
{
    SCIM_DEBUG_IMENGINE(2) << "focus_out.\n";
    reset ();
}

void
HangulInstance::trigger_property (const String &property)
{
    String hangul_prop = property.substr (property.find_last_of ('/') + 1);

    SCIM_DEBUG_IMENGINE(2) << "trigger_property.\n";

    if (property == SCIM_PROP_INPUT_MODE_LATIN) {
	set_input_mode (INPUT_MODE_DIRECT);
    } else if (property == SCIM_PROP_INPUT_MODE_HANGUL) {
	set_input_mode (INPUT_MODE_HANGUL);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_2BUL) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_2BUL);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_3BUL_2BUL_SHIFTED) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_2BUL_SHIFTED);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_3BUL_390) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_390);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_3BUL_FINAL) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_FINAL);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_3BUL_NO_SHIFT) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_NO_SHIFT);
    } else if (property == SCIM_PROP_INPUT_LAYOUT_3BUL_YETGEUL) {
	set_input_layout (SCIM_HANGUL_INPUT_LAYOUT_3BUL_YETGEUL);
    }
}

static int
get_index_of_candidate_table (ucs4_t ch)
{
    int first, last, mid;

    /* binary search */
    first = 0;
    last = sizeof (candidate_table)/sizeof (candidate_table [0]) - 1;

    while (first <= last) {
        mid = (first + last) / 2;

        if (ch == candidate_table[mid][0].ch)
            return mid;

        if (ch < candidate_table[mid][0].ch)
            last = mid - 1;
        else
            first = mid + 1;
    }

    return -1;
}

void
HangulInstance::update_candidates ()
{
    WideString wstr = get_preedit_string ();
    if (wstr.empty())
        return; 

    ucs4_t ch = wstr[0];

    m_lookup_table.clear ();
    m_candidate_comments.clear ();

    if (ch > 0) {
        int index = get_index_of_candidate_table (ch);
        if (index != -1) {
            for (const CandidateItem *it = candidate_table [index] + 1; it->ch; ++it) {
                m_lookup_table.append_candidate (it->ch);
                m_candidate_comments.push_back (String (it->comment));
            }

            m_lookup_table.set_page_size (9);
            m_lookup_table.show_cursor ();

            update_lookup_table (m_lookup_table);
            show_lookup_table ();

            hangul_update_aux_string ();
        }
    }
}

void
HangulInstance::delete_candidates ()
{
    m_lookup_table.clear ();
    m_candidate_comments.clear ();
    hide_lookup_table ();
    hide_aux_string ();
}

void
HangulInstance::hangul_update_aux_string ()
{
    if (!m_factory->m_show_candidate_comment || !m_lookup_table.number_of_candidates ()) {
        hide_aux_string ();
        return;
    }

    size_t cursor = m_lookup_table.get_cursor_pos ();

    if (cursor >= m_candidate_comments.size () || !m_candidate_comments [cursor].length ()) {
        hide_aux_string ();
        return;
    }

    update_aux_string (m_lookup_table.get_candidate (cursor) + utf8_mbstowcs (String (" : ") + m_candidate_comments [cursor]));
    show_aux_string ();
}

void
HangulInstance::hangul_update_preedit_string ()
{
    WideString wstr = get_preedit_string ();

    if (wstr.length ()) {
        AttributeList attrs;
        attrs.push_back (Attribute (0, wstr.length (), SCIM_ATTR_DECORATE, SCIM_ATTR_DECORATE_REVERSE));
        show_preedit_string ();
        update_preedit_string (wstr, attrs);
        update_preedit_caret (wstr.length());
    } else {
        hide_preedit_string ();
    }
}

bool
HangulInstance::match_key_event (const KeyEventList &keys, const KeyEvent &key) const
{
    KeyEventList::const_iterator kit; 

    for (kit = keys.begin (); kit != keys.end (); ++kit) {
        if (key.code == kit->code && key.mask == kit->mask)
            if (!(key.mask & SCIM_KEY_ReleaseMask) || m_prev_key.code == key.code)
                return true;
    }
    return false;
}

void
HangulInstance::install_properties (void)
{
    if (m_properties.size () <= 0) {
        Property prop;

	prop = Property (SCIM_PROP_INPUT_MODE,
			 "\xED\x95x9C", String (""), _("Input mode"));
	m_properties.push_back (prop);

	prop = Property (SCIM_PROP_INPUT_MODE_LATIN,
			 _("Latin"), String (""), _("Direct input"));
	m_properties.push_back (prop);

	prop = Property (SCIM_PROP_INPUT_MODE_HANGUL,
			 _("Hangul"), String (""), _("Hangul"));
	m_properties.push_back (prop);

	if (!m_factory->m_hide_layout) {
	    prop = Property (SCIM_PROP_INPUT_LAYOUT,
			     "2bul", String (""), _("Input Layout"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_2BUL,
			     _("2bul"), String (""), _("2bul"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_3BUL_2BUL_SHIFTED,
			     _("3bul 2bul-shifted"), String (""), _("3bul 2bul-shifted"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_3BUL_390,
			     _("3bul 390"), String (""), _("3bul 390"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_3BUL_FINAL,
			     _("3bul Final"), String (""), _("3bul Final"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_3BUL_NO_SHIFT,
			     _("3bul No-Shift"), String (""), _("3bul No-Shift"));
	    m_properties.push_back (prop);

	    prop = Property (SCIM_PROP_INPUT_LAYOUT_3BUL_YETGEUL,
			     _("3bul Yetgeul"), String (""), _("3bul Yetgeul"));
	    m_properties.push_back (prop);
	}
    }

    set_input_layout (get_input_layout ());
    if (m_on_off_state)
	set_input_mode (INPUT_MODE_HANGUL);
    else
	set_input_mode (INPUT_MODE_DIRECT);

    register_properties (m_properties);
}

void
HangulInstance::set_input_mode (HangulInputMode mode)
{
    const char *label = "";

    switch (mode) {
	case INPUT_MODE_DIRECT:
	    label = "_A";
	    break;
	case INPUT_MODE_HANGUL:
	    label = "\xED\x95\x9C";
	    break;
	case INPUT_MODE_HANJA:
	    /* FIXME */
	    break;
	default:
	    break;
    }

    if (label && *label) {
	PropertyList::iterator it = std::find (m_properties.begin (),
					       m_properties.end (),
					       SCIM_PROP_INPUT_MODE);
	if (it != m_properties.end ()) {
		it->set_label (label);
		update_property (*it);
	}
    }
    m_input_mode = mode;
    if (mode == INPUT_MODE_DIRECT) {
        m_on_off_state = false;
    } else {
        m_on_off_state = true;
    }
}

HangulInputMode
HangulInstance::get_input_mode (void)
{
    return m_input_mode;
}

void
HangulInstance::set_input_layout (InputLayout layout)
{
    const char *label = "";
    HangulKeyboardType keyboard = HANGUL_KEYBOARD_2;

    /* FIXME: better label to keep the same width.
     *        flapping toolbar every input layout change
     *        may be annoying.
     */
    switch (layout) {
	case SCIM_HANGUL_INPUT_LAYOUT_2BUL:
	    label = "2bul";
	    keyboard = HANGUL_KEYBOARD_2;
	    break;
	case SCIM_HANGUL_INPUT_LAYOUT_3BUL_2BUL_SHIFTED:
	    label = "3bul 2bul-shifted";
	    keyboard = HANGUL_KEYBOARD_32;
	    break;
	case SCIM_HANGUL_INPUT_LAYOUT_3BUL_390:
	    label = "3bul 390";
	    keyboard = HANGUL_KEYBOARD_390;
	    break;
	case SCIM_HANGUL_INPUT_LAYOUT_3BUL_FINAL:
	    label = "3bul Final";
	    keyboard = HANGUL_KEYBOARD_3FINAL;
	    break;
	case SCIM_HANGUL_INPUT_LAYOUT_3BUL_NO_SHIFT:
	    label = "3bul No-Shift";
	    keyboard = HANGUL_KEYBOARD_3NOSHIFT;
	    break;
	case SCIM_HANGUL_INPUT_LAYOUT_3BUL_YETGEUL:
	    label = "3bul Yetgeul";
	    keyboard = HANGUL_KEYBOARD_3YETGUL;
	    break;
	default:
	    break;
    }

    if (label && *label) {
	PropertyList::iterator it = std::find (m_properties.begin (),
					       m_properties.end (),
					       SCIM_PROP_INPUT_LAYOUT);
	if (it != m_properties.end ()) {
		it->set_label (_(label));
		update_property (*it);
	}
    }
    if (layout != get_input_layout ()) {
	m_input_layout = layout;
	hangul_ic_set_keyboard(m_hic, keyboard);
    }
}

InputLayout
HangulInstance::get_input_layout (void)
{
    return m_input_layout;
}

//vim: ts=4:nowrap:ai:expandtab:
