/** @file scim_hangul_imengine.h
 */

/* 
 * Smart Common Input Method
 * 
 * Copyright (C) 2004-2006 Choe Hwanjin
 * Copyright (c) 2004-2006 James Su <suzhe@tsinghua.org.cn>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: scim_hangul_imengine.h,v 1.3 2006/01/17 02:17:01 suzhe Exp $
 */

#if !defined (__SCIM_HANGUL_IMENGINE_H)
#define __SCIM_HANGUL_IMENGINE_H

#include "hangul.h"

using namespace scim;

enum HangulInputMode
{
    INPUT_MODE_DIRECT,
    INPUT_MODE_HANGUL,
    INPUT_MODE_HANJA
};

enum HangulOutputMode
{
    OUTPUT_MODE_SYLLABLE = 0,
    OUTPUT_MODE_JAMO     = 1 << 1,
    OUTPUT_MODE_JAMO_EXT = 1 << 2
};

struct HangulCombination 
{
  uint32  key;
  ucs4_t  code;
};

struct HangulFactoryData
{
    const char              *uuid;
    const char              *name;

    HangulKeyboardType       keyboard_type;
};

class HangulInstance;

class HangulFactory : public IMEngineFactoryBase
{
    String                   m_uuid;
    String                   m_input_layout;

    HangulKeyboardType       m_keyboard_type;

    bool                     m_always_use_jamo;

    bool                     m_show_candidate_comment;
    bool                     m_use_ascii_mode;
    bool                     m_hide_layout;
    bool                     m_lookup_table_vertical;

    KeyEventList             m_trigger_keys;
    KeyEventList             m_hangul_hanja_keys;
    KeyEventList             m_on_off_keys;

    friend class HangulInstance;

public:
    ConfigPointer m_config;

    HangulFactory (const ConfigPointer &config, const char *uuid);

    virtual ~HangulFactory ();

    virtual WideString  get_name () const;
    virtual WideString  get_authors () const;
    virtual WideString  get_credits () const;
    virtual WideString  get_help () const;
    virtual String      get_uuid () const;
    virtual String      get_icon_file () const;

    virtual IMEngineInstancePointer create_instance (const String& encoding, int id = -1);

    virtual void append_config_listener (HangulInstance *listener);
    virtual void remove_config_listener (HangulInstance *listener);

private:
    Connection                   m_reload_signal_connection;
    std::vector<HangulInstance*> m_config_listeners;

    void reload_config (const ConfigPointer &config);
};

class HangulInstance : public IMEngineInstanceBase
{
    HangulFactory       *m_factory;

    CommonLookupTable    m_lookup_table;
    std::vector<String>  m_candidate_comments;

    KeyEvent             m_prev_key;

    HangulInputContext  *m_hic;

    int                  m_index;         /* stack index */
    ucs4_t               m_stack     [12];

    int                  m_lindex;        /* leading consonant index */
    int                  m_vindex;        /* vowel index */
    int                  m_tindex;        /* trailing consonant index */
    ucs4_t               m_choseong  [4];
    ucs4_t               m_jungseong [4];
    ucs4_t               m_jongseong [4];

    HangulInputMode      m_input_mode;
    int                  m_output_mode;

    bool (HangulInstance::*m_hangul_composer) (const KeyEvent &key);

    bool                 m_on_init;
    bool                 m_on_off_state;
    InputLayout          m_input_layout;

    PropertyList         m_properties;

public:
    HangulInstance (HangulFactory *factory,
                    const String  &encoding,
                    int            id = -1);

    virtual ~HangulInstance ();

    virtual bool process_key_event (const KeyEvent& key);
    virtual void move_preedit_caret (unsigned int pos);
    virtual void select_candidate (unsigned int item);
    virtual void update_lookup_table_page_size (unsigned int page_size);
    virtual void lookup_table_page_up ();
    virtual void lookup_table_page_down ();
    virtual void reset ();
    virtual void focus_in ();
    virtual void focus_out ();
    virtual void trigger_property (const String &property);
    virtual void reload_config    (const ConfigPointer &config);

private:
    bool is_backspace_key (const KeyEvent &key) const {
        return (key.code == SCIM_KEY_BackSpace);
    }

    bool is_hangul_hanja_key (const KeyEvent &key) const {
        return match_key_event (m_factory->m_hangul_hanja_keys, key);
    }

    bool is_on_off_key (const KeyEvent &key) const {
        return match_key_event (m_factory->m_on_off_keys, key);
    }

    void toggle_on_off_state (void) {
        if (m_on_off_state)
	    set_input_mode (INPUT_MODE_DIRECT);
	else
            set_input_mode (INPUT_MODE_HANGUL);
    }

    bool is_ime_activated (void) {
        return m_on_off_state;
    }

    /* preedit string */
    WideString get_preedit_string () {
        WideString wstr;
        const wchar_t *wcs = hangul_ic_get_preedit_string(m_hic);
        while (*wcs != L'\0')
            wstr.push_back (*wcs++);
        return wstr;
    }

    WideString get_commit_string () {
        WideString wstr;
        const wchar_t *wcs = hangul_ic_get_commit_string(m_hic);
        while (*wcs != L'\0')
            wstr.push_back (*wcs++);
        return wstr;
    }

    void hangul_update_preedit_string ();

    /* aux string */
    void hangul_update_aux_string ();

    /* candidate functions */
    void   update_candidates ();
    void   delete_candidates ();

    /* candidate keys */
    bool   candidate_key_event (const KeyEvent &key);

    /* match key event */
    bool   match_key_event (const KeyEventList &keys, const KeyEvent &key) const;

    void install_properties (void);

    void        set_input_layout (InputLayout layout);
    InputLayout get_input_layout (void);

    void            set_input_mode (HangulInputMode mode);
    HangulInputMode get_input_mode (void);

    void cursor_up_on_lookup_table   (void);
    void cursor_down_on_lookup_table (void);
    void page_up_on_lookup_table     (void);
    void page_down_on_lookup_table   (void);
};
#endif

/*
vi:ts=4:nowrap:ai:expandtab
*/
